﻿// Le Transit-O-Tron : comment ameliorer son transit en parapente
//
//===========================================================================================
// Copyright (c) 2011 Thierry Bernier
// A utiliser a tes risques et périls : ton disque dur pourrait bien prendre feu.
// Tu peux utiliser ce code source comme bon te semble, du moment que ce copyright
// est conserve.
//===========================================================================================

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Windows.Forms.DataVisualization.Charting;


namespace Transit_o_Tron
{
    public partial class Form1 : Form
    {
        public Form1 ()
        {
            InitializeComponent();

            // Le dessin de la trajectoire
            mTrajet = new Dessin ();
            mTrajet.Paint += new PaintEventHandler (OnPaintTrajet);
            mTrajet.MouseDown += new MouseEventHandler (OnMouseDown);
            //mTrajet.MouseMove += new MouseEventHandler (OnMouseMove);
            //mTrajet.MouseUp += new MouseEventHandler (OnMouseUp);
            //mTrajet.MouseWheel += new MouseEventHandler (OnMouseWheel);
            mTrajet.BackColor = Color.LightBlue;
            mTrajet.Dock = DockStyle.Fill;
            splitContainer1.Panel1.Controls.Add (mTrajet);

            Series s;
            // Le dessin de la polaire
            chartAile.ChartAreas[0].AxisX.Minimum = 15.0;
            chartAile.ChartAreas[0].AxisX.Maximum = 65.0;
            chartAile.ChartAreas[0].AxisY.Minimum = -2.5;
            chartAile.ChartAreas[0].AxisY.Maximum = -0.5;
            chartAile.ChartAreas[0].AxisX.Interval = 5.0;
            chartAile.ChartAreas[0].AxisY.Interval = 0.5;
            chartAile.Legends[0].Enabled = false;
            s = new Series ();
            s.ChartType = SeriesChartType.Point;
            s.IsValueShownAsLabel = false;
            s.Points.AddXY (24.0, -1.9);
            s.Points.AddXY (31.0, -0.9);
            s.Points.AddXY (39.0, -1.1);
            s.Points.AddXY (57.0, -1.9);
            chartAile.Series.Add (s);

            //chartPilote.ChartAreas[0].AxisY.Title
            chartPilote.ChartAreas[0].AxisX.Minimum = -3.0;
            chartPilote.ChartAreas[0].AxisX.Maximum = 3.0;
            chartPilote.ChartAreas[0].AxisY.Minimum = -1.0;
            chartPilote.ChartAreas[0].AxisY.Maximum = 1.0;
            chartPilote.ChartAreas[0].AxisX.Interval = 0.5;
            chartPilote.ChartAreas[0].AxisY.Interval = 0.5;
            chartPilote.Legends[0].Enabled = false;

            // Reaction du pilote au taux de chute
            s = new Series ();
            s.ChartType = SeriesChartType.Point;
            s.LegendText = "TxChute";
            s.IsValueShownAsLabel = false;
            s.Points.AddXY (-2.5,  0.9);
            s.Points.AddXY (-1.0,  0.1);
            s.Points.AddXY ( 0.0, -0.5);
            s.Points.AddXY ( 1.0, -0.5);
            s.Points.AddXY ( 2.5, -0.9);
            s.ToolTip = "Réaction au taux de chute";
            chartPilote.Series.Add(s);                  // Index 0
            //chartPilote.Legends[0].Title = "TxChute";

            // Reaction du pilote L/D
            //s = new Series ();
            //s.ChartType = SeriesChartType.Spline;
            //s.LegendText = "Finesse sol";
            ////s.XAxisType = AxisType.Primary;
            //s.IsValueShownAsLabel = false;
            //s.Points.AddXY (1.0, 1.0);
            //s.Points.AddXY (2.0, 1.0);
            //s.Points.AddXY (3.0, 1.0);
            //s.Points.AddXY (4.0, 0.8);
            //s.Points.AddXY (5.0, 0.7);
            //s.Points.AddXY (6.0, 0.6);
            //s.Points.AddXY (7.0, 0.5);
            //s.Points.AddXY (8.0, 0.1);
            //s.Points.AddXY (9.0, 0.0);
            //s.Points.AddXY (10.0, 0.0);
            //s.Points.AddXY (11.0, -0.5);
            //s.Points.AddXY (12.0, -0.7);
            //s.Points.AddXY (13.0, -0.8);
            //s.ToolTip = "Réaction à la finesse sol";
            //chartPilote.Series.Add (s);                  // Index 1

            // Le modele
            pilote = chartPilote_Value ();
            aile   = chartAile_Value ();
            ciel   = new Modele.Sky ();

            ciel.mVent.X = (double)numericUpDownGvh.Value / 3.6;
            altStart = (double)numericUpDownHdep.Value;

            trajet = ciel.Fly (altStart, pilote, aile);
        }

        // Le modele
        private Modele.Pilote pilote;
        private Modele.Aile aile;
        private Modele.Sky ciel;
        private Modele.Trajet trajet;
        private double altStart;

        // L'IHM
        private Random rnd = new Random ();
        private Dessin mTrajet;
        //private Dessin mPolaire;
        private Rule[] mRules = new Rule[4];
        Font mFont = new Font ("Arial", 9);
        Pen mPenBlk = new Pen (Color.Black);
        Pen mPenGrn = new Pen (Color.Green);
        Pen mPenRed = new Pen (Color.Red);

        // Echelle de dessin
        private double mXmin, mXmax, mHmin, mHmax, mKx, mKy;
        private const int marginGround = 10;


        // Une surface de dessin
        private class Dessin : UserControl
        {
            public Dessin ()
            {
                ResizeRedraw = true;
            }

            // SetStyle (UserPaint|AllPaintingInWmPaint) marche pas ?
//            protected override void OnPaintBackground (PaintEventArgs e)
//            {
//            }
        };

        private void SetScale ()
        {
            // Le domaine rendu est cale sur le max, pour voir la trajectoire se deplacer
            //mXmin = (double) numericUpDownZle.Minimum * 1000.0d;
            //mXmax = (double) numericUpDownZri.Maximum * 1000.0d;
            //mHmin = (double) numericUpDownZlo.Minimum * 1.0d;
            //mHmax = (double) numericUpDownZhi.Maximum * 1.0d;
            mXmin = Modele.Sky.clipMinX;
            mXmax = Modele.Sky.clipMaxX + 500.0;
            mHmin = Modele.Sky.clipMinH;
            mHmax = Modele.Sky.clipMaxH;
            Size pix = mTrajet.ClientSize;
            mKx = (double) pix.Width  / (mXmax - mXmin);
            mKy = (double) (pix.Height-marginGround) / (mHmax - mHmin);
        }

        private int ToPixelX (double x)
        {
            return (int)((double)(x - mXmin) * mKx);
        }
        private int ToPixelY (double h)
        {
            return (int)((double)(mHmax - h) * mKy);
        }
        private int ToValueX (int x)
        {
            return (int) ((double) x / mKx + mXmin);
        }
        private int ToValueH (int y)
        {
            return (int) ((double) mHmax - y / mKy);
        }

        private void OnPaintTrajet (object sender, PaintEventArgs e)
        {
            //Trace.WriteLine ("OnPaintTrajet " + e.ClipRectangle);
            base.OnPaint (e);
            SetScale ();

            // Le decor
            e.Graphics.FillRectangle (Brushes.Chocolate, 0, mTrajet.ClientSize.Height-marginGround, mTrajet.ClientSize.Width, marginGround);

            // Les thermiques
            foreach (Modele.Zone z in ciel.mZones)
            {
                int x = ToPixelX(z.Xlole);
                int y = ToPixelY(z.Hhi);
                e.Graphics.DrawRectangle ((z == selectedZone) ? mPenRed : mPenGrn,
                                           x, y, ToPixelX (z.Xlori) - x, ToPixelY (z.Hlo)-y);
            }

            // La trajectoire
            Modele.Mobile lm = null;
            int lstx = -999;
            int lsty = -999;
            foreach (Modele.Mobile m in trajet.mPositions)
            {
                int curx = ToPixelX (m.P.X);
                int cury = ToPixelY (m.P.H);
                e.Graphics.DrawRectangle (mPenBlk, curx - 2, cury - 2, 4, 4);
                if (lstx != -999) e.Graphics.DrawLine (mPenBlk, curx, cury, lstx, lsty);
                lstx = curx;
                lsty = cury;
                lm = m;
            }

            // La date et distance de fin
            if (lm != null)
            {
                String txt = ((int) lm.P.X) + " m\n" + ((int) lm.T) + " s";
                e.Graphics.DrawString (txt, mFont, Brushes.Black, lstx, lsty-30);
            }

        }

        private Modele.Zone selectedZone = null;
        private void OnMouseDown (object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                return;
            }
            if (e.Button == MouseButtons.Left)
            {
                Modele.Point p = new Modele.Point (ToValueX (e.X), ToValueH (e.Y));
                selectedZone = ciel.ZoneContaining (p);
                if (selectedZone == null)
                    buttonZone.Text = "Ajouter";
                else
                {
                    buttonZone.Text = "Supprimer";
                    numericUpDownZle.Value = (decimal)selectedZone.Xlole / 1000.0m;
                    numericUpDownZri.Value = (decimal)selectedZone.Xlori / 1000.0m;
                    numericUpDownZlo.Value = (decimal)selectedZone.Hlo / 1.0m;
                    numericUpDownZhi.Value = (decimal)selectedZone.Hhi / 1.0m;
                    numericUpDownZvh.Value = (decimal)selectedZone.Vent.X * 3.6m;
                    numericUpDownZvv.Value = (decimal)selectedZone.Vent.H / 1.0m;
                }
                mTrajet.Invalidate ();
            }
        }


        //private void OnPaintPolaire (object sender, PaintEventArgs e)
        //{
        //    Trace.WriteLine ("OnPaintPolaire " + e.ClipRectangle);
        //    base.OnPaint (e);
        //}

        private void RefreshTrajet ()
        {
            trajet = ciel.Fly (altStart, pilote, aile);
            mTrajet.Invalidate ();
        }

        private void SetZone (Modele.Zone z)
        {
            z.Xlole = z.Xhile = (double)numericUpDownZle.Value * 1000.0d;
            z.Xlori = z.Xhiri = (double)numericUpDownZri.Value * 1000.0d;
            z.Hlo = (double)numericUpDownZlo.Value * 1.0d;
            z.Hhi = (double)numericUpDownZhi.Value * 1.0d;
            z.Vent.X = (double)numericUpDownZvh.Value / 3.6;
            z.Vent.H = (double)numericUpDownZvv.Value * 1.0d;
        }

        private void buttonZone_Click (object sender, EventArgs e)
        {
            if (selectedZone == null)       // Ajouter un thermique
            {
                Modele.Zone z = new Modele.Zone (0.0, 0.0, 0.0, 0.0);
                SetZone (z);
                ciel.mZones.Add (z);
            }
            else                            // Supprimer ce thermique
            {
                ciel.mZones.Remove (selectedZone);
                selectedZone = null;
                buttonZone.Text = "Ajouter";
            }
            RefreshTrajet ();
        }

        private void numericUpDown8_ValueChanged (object sender, EventArgs e)
        {
            ciel.mVent.X = (double)numericUpDownGvh.Value / 3.6;
            RefreshTrajet ();
        }

        private void numericUpDownHdep_ValueChanged (object sender, EventArgs e)
        {
            altStart = (double)numericUpDownHdep.Value;
            RefreshTrajet ();
        }


        private Modele.Aile chartAile_Value ()
        {
            Modele.Speed[] spd = new Modele.Speed[chartAile.Series[0].Points.Count];
            for (int i = 0; i < chartAile.Series[0].Points.Count; ++i)
                spd[i] = new Modele.Speed (chartAile.Series[0].Points[i].XValue / 3.6,
                                           chartAile.Series[0].Points[i].YValues[0]);
            return new Modele.Aile (spd);
        }

        private DataPoint selectedDataPoint = null;
        private int selectedDataIndex;

        private void chartAile_MouseDown (object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                if (chartAile.Series[0].ChartType == SeriesChartType.Spline)
                    chartAile.Series[0].ChartType = SeriesChartType.Point;
                else
                    chartAile.Series[0].ChartType = SeriesChartType.Spline;
                chartAile.Invalidate ();
                return;
            }
            // Debug
            if (e.Button == MouseButtons.Middle)
            {
                Series s = new Series ();
                s.ChartType = SeriesChartType.Line;
                for (int i = -10; i <= 10; ++i)
                {
                    Modele.Speed spd = aile.AirSpeed ((double) i / 10.0);
                    s.Points.AddXY (spd.X * 3.6, spd.H);
                }
                chartAile.Series.Add(s);
            }


            // Call Hit Test Method
            HitTestResult hitResult = chartAile.HitTest (e.X, e.Y);

            // Initialize currently selected data point
            selectedDataPoint = null;
            if (hitResult.ChartElementType == ChartElementType.DataPoint)
            {
                selectedDataPoint = (DataPoint)hitResult.Object;
                selectedDataIndex = hitResult.PointIndex;

                // Show point value as label
                selectedDataPoint.IsValueShownAsLabel = true;

                // Set cursor shape
                chartAile.Cursor = Cursors.SizeNS;
            }

        }

        private void chartAile_MouseMove (object sender, MouseEventArgs e)
        {
            // Check if data point selected
            if (selectedDataPoint != null)
            {
                // Mouse coordinates should not be outside of the chart 
                int coordinate = e.Y;
                if (coordinate < 0) coordinate = 0;
                if (coordinate > chartAile.Size.Height - 1) coordinate = chartAile.Size.Height - 1;

                // Calculate new Y value from current cursor position
                double yValue = chartAile.ChartAreas[0].AxisY.PixelPositionToValue (coordinate);
                yValue = Math.Min (yValue, chartAile.ChartAreas[0].AxisY.Maximum);
                yValue = Math.Max (yValue, chartAile.ChartAreas[0].AxisY.Minimum);

                // Update selected point Y value
                selectedDataPoint.YValues[0] = yValue;

                // Mouse coordinates should not be outside of the chart 
                coordinate = e.X;
                if (coordinate < 0) coordinate = 0;
                if (coordinate > chartAile.Size.Width - 1) coordinate = chartAile.Size.Width - 1;

                // Calculate new X value from current cursor position
                double xValue = chartAile.ChartAreas[0].AxisX.PixelPositionToValue (coordinate);
                xValue = Math.Min (xValue, chartAile.ChartAreas[0].AxisX.Maximum);
                xValue = Math.Max (xValue, chartAile.ChartAreas[0].AxisX.Minimum);

                // Les X des points doivent rester croissants
                if (selectedDataIndex > 0)
                    xValue = Math.Max (xValue, chartAile.Series[0].Points[selectedDataIndex - 1].XValue + 0.5);
                if (selectedDataIndex < 3)
                    xValue = Math.Min (xValue, chartAile.Series[0].Points[selectedDataIndex + 1].XValue - 0.5);

                // Update selected point X value
                selectedDataPoint.XValue = xValue;

                // Invalidate chart
                chartAile.Invalidate ();
            }
            else
            {
                // Set different shape of cursor over the data points
                HitTestResult hitResult = chartAile.HitTest (e.X, e.Y);
                if (hitResult.ChartElementType == ChartElementType.DataPoint)
                {
                    chartAile.Cursor = Cursors.Hand;
                }
                else
                {
                    chartAile.Cursor = Cursors.Default;
                }
            }

        }

        private void chartAile_MouseUp (object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Middle)
            {
                chartAile.Series.RemoveAt (1);
                return;
            }

            // Initialize currently selected data point
            if (selectedDataPoint != null)
            {
                // Hide point label
                selectedDataPoint.IsValueShownAsLabel = false;

                // reset selected object
                selectedDataPoint = null;

                aile = chartAile_Value ();
                RefreshTrajet ();

                // Invalidate chart
                chartAile.Invalidate ();

                // Reset cursor style
                chartAile.Cursor = Cursors.Default;
            }
        }

        private void buttonExemple_Click (object sender, EventArgs e)
        {
            Modele.Zone z;

            ciel.mZones.Clear ();

            z = new Modele.Zone (1000.0, 1300.0, 100.0, 900.0);
            z.Vent.X = -20.0 / 3.6;
            z.Vent.H = 3.0;
            ciel.mZones.Add (z);

            z = new Modele.Zone (1400.0, 1600.0, 100.0, 600.0);
            z.Vent.X = -20.0 / 3.6;
            z.Vent.H = -6.0;
            ciel.mZones.Add (z);

            z = new Modele.Zone (2400.0, 2800.0, 400.0, 1600.0);
            z.Vent.X = 10.0 / 3.6;
            z.Vent.H = 6.0;
            ciel.mZones.Add (z);

            RefreshTrajet ();

        }

        private void numericUpDownZ_ValueChanged (object sender, EventArgs e)
        {
            if (selectedZone != null)
            {
                //SetZone (selectedZone);
                RefreshTrajet ();
            }
        }


        private Modele.Pilote chartPilote_Value ()
        {
            Modele.Speed[] speed = new Modele.Speed[chartPilote.Series[0].Points.Count];
            double[] cmd = new double[chartPilote.Series[0].Points.Count];

            for(int i = 0; i < chartPilote.Series[0].Points.Count; ++i)
            {
                speed[i] = new Modele.Speed (0.0, chartPilote.Series[0].Points[i].XValue);
                cmd[i] = chartPilote.Series[0].Points[i].YValues[0];
            };
            return new Modele.Pilote (speed, cmd);
        }


        private void chartPilote_MouseDown (object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                if (chartPilote.Series[0].ChartType == SeriesChartType.Spline)
                    chartPilote.Series[0].ChartType = SeriesChartType.Point;
                else
                    chartPilote.Series[0].ChartType = SeriesChartType.Spline;
                chartPilote.Invalidate ();
                return;
            }
            // Debug
            if (e.Button == MouseButtons.Middle)
            {
                Series s = new Series ();
                Modele.Speed spd = new Modele.Speed (0.0, 0.0);
                s.ChartType = SeriesChartType.Line;
                for (int i = -30; i <= 30; i += 2)
                {
                    spd.H = (double) i / 10.0;
                    double cmd = pilote.Commande (spd);
                    s.Points.AddXY (spd.H, cmd);
                }
                chartPilote.Series.Add (s);
            }

            // Call Hit Test Method
            HitTestResult hitResult = chartPilote.HitTest (e.X, e.Y);

            // Initialize currently selected data point
            selectedDataPoint = null;
            if (hitResult.ChartElementType == ChartElementType.DataPoint)
            {
                selectedDataPoint = (DataPoint)hitResult.Object;
                selectedDataIndex = hitResult.PointIndex;

                // Show point value as label
                selectedDataPoint.IsValueShownAsLabel = true;

                // Set cursor shape
                chartPilote.Cursor = Cursors.SizeNS;
            }
        }

        private void chartPilote_MouseMove (object sender, MouseEventArgs e)
        {
            // Check if data point selected
            if (selectedDataPoint != null)
            {
                // Mouse coordinates should not be outside of the chart 
                int coordinate = e.Y;
                if (coordinate < 0) coordinate = 0;
                if (coordinate > chartPilote.Size.Height - 1) coordinate = chartPilote.Size.Height - 1;

                // Calculate new Y value from current cursor position
                double yValue = chartPilote.ChartAreas[0].AxisY.PixelPositionToValue (coordinate);
                yValue = Math.Min (yValue, chartPilote.ChartAreas[0].AxisY.Maximum);
                yValue = Math.Max (yValue, chartPilote.ChartAreas[0].AxisY.Minimum);

                // Update selected point Y value
                selectedDataPoint.YValues[0] = yValue;

                // Mouse coordinates should not be outside of the chart 
                coordinate = e.X;
                if (coordinate < 0) coordinate = 0;
                if (coordinate > chartPilote.Size.Width - 1) coordinate = chartPilote.Size.Width - 1;

                // Calculate new X value from current cursor position
                double xValue = chartPilote.ChartAreas[0].AxisX.PixelPositionToValue (coordinate);
                xValue = Math.Min (xValue, chartPilote.ChartAreas[0].AxisX.Maximum);
                xValue = Math.Max (xValue, chartPilote.ChartAreas[0].AxisX.Minimum);

                // Les X des points doivenbt rester croissants
                if (selectedDataIndex > 0)
                    xValue = Math.Max (xValue, chartPilote.Series[0].Points[selectedDataIndex - 1].XValue + 0.1);
                if (selectedDataIndex < 3)
                    xValue = Math.Min (xValue, chartPilote.Series[0].Points[selectedDataIndex + 1].XValue - 0.1);

                // Update selected point X value
                selectedDataPoint.XValue = xValue;

                // Invalidate chart
                chartPilote.Invalidate ();
            }
            else
            {
                // Set different shape of cursor over the data points
                HitTestResult hitResult = chartPilote.HitTest (e.X, e.Y);
                if (hitResult.ChartElementType == ChartElementType.DataPoint)
                {
                    chartPilote.Cursor = Cursors.Hand;
                }
                else
                {
                    chartPilote.Cursor = Cursors.Default;
                }
            }
        }

        private void chartPilote_MouseUp (object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Middle)
            {
                chartPilote.Series.RemoveAt (1);
                return;
            }
            // Initialize currently selected data point
            if (selectedDataPoint != null)
            {
                // Hide point label
                selectedDataPoint.IsValueShownAsLabel = false;

                // reset selected object
                selectedDataPoint = null;

                pilote = chartPilote_Value ();
                RefreshTrajet ();

                // Invalidate chart
                chartPilote.Invalidate ();

                // Reset cursor style
                chartPilote.Cursor = Cursors.Default;
            }
        }

    }
}
